#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <stdarg.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#include "wmgeneral.h"

Window Root;
int screen;
int x_fd;
int d_depth;
XSizeHints mysizehints;
XWMHints mywmhints;
Pixel back_pix, fore_pix;
char *Geometry = "";
Window iconwin, win;
GC NormalGC, BlackGC, DarkerGC;
XpmIcon wmgen;
Pixmap pixmask;

typedef struct {
  int enable;
  int top;
  int bottom;
  int left;
  int right;
} MOUSE_REGION;

MOUSE_REGION mouse_region[MAX_MOUSE_REGION];

static void GetXPM(XpmIcon *wmgen, const char * const pixmap_bytes[])
{
  XWindowAttributes attributes;
  int err;
  XGetWindowAttributes(display, Root, &attributes);
  wmgen->attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  err = XpmCreatePixmapFromData(display, Root, (char **)pixmap_bytes,
    &(wmgen->pixmap), &(wmgen->mask), &(wmgen->attributes));
  if (err != XpmSuccess) {
    fprintf(stderr, "Not enough free colorcells.\n");
    exit(1);
  }
}

static Pixel GetColor(char *name)
{
  XColor color;
  XWindowAttributes attributes;
  XGetWindowAttributes(display, Root, &attributes);
  color.pixel = 0;
  if (!XParseColor(display, attributes.colormap, name, &color)) {
    fprintf(stderr, "wm.app: can't parse %s.\n", name);
  } else if (!XAllocColor(display, attributes.colormap, &color)) {
    fprintf(stderr, "wm.app: can't allocate %s.\n", name);
  }
  return color.pixel;
}

static int flush_expose(Window w)
{
  XEvent dummy;
  int i=0;
  while (XCheckTypedWindowEvent(display, w, Expose, &dummy))
    i++;
  return i;
}

void RedrawWindow(void)
{
  flush_expose(iconwin);
  XCopyArea(display, wmgen.pixmap, iconwin, NormalGC, 
    0, 0, wmgen.attributes.width, wmgen.attributes.height, 0, 0);
  flush_expose(win);
  XCopyArea(display, wmgen.pixmap, win, NormalGC,
    0, 0, wmgen.attributes.width, wmgen.attributes.height, 0, 0);
}

void AddMouseRegion(int index, int left, int top, int right, int bottom)
{
  if (index < MAX_MOUSE_REGION) {
    mouse_region[index].enable = 1;
    mouse_region[index].top = top;
    mouse_region[index].left = left;
    mouse_region[index].bottom = bottom;
    mouse_region[index].right = right;
  }
}

int CheckMouseRegion(int x, int y)
{
  int i;
  int found;
  found = 0;
  for (i=0; i<MAX_MOUSE_REGION && !found; i++) {
    if (mouse_region[i].enable &&
        x <= mouse_region[i].right &&
        x >= mouse_region[i].left &&
        y <= mouse_region[i].bottom &&
        y >= mouse_region[i].top)
      found = 1;
  }
  if (!found) return -1;
    return (i-1);
}

void copyXPMArea(int x, int y, int sx, int sy, int dx, int dy)
{
  XCopyArea(display, wmgen.pixmap, wmgen.pixmap, NormalGC,
    x, y, sx, sy, dx, dy);
}

void copyXBMArea(int x, int y, int sx, int sy, int dx, int dy)
{
  XCopyArea(display, wmgen.mask, wmgen.pixmap, NormalGC,
    x, y, sx, sy, dx, dy);
}

void drawline(int x1, int y1, int x2, int y2)
{
  XDrawLine(display, wmgen.pixmap, NormalGC, x1, y1, x2, y2);
}

void darkline(int x1, int y1, int x2, int y2)
{
  XDrawLine(display, wmgen.pixmap, DarkerGC, x1, y1, x2, y2);
}

void cleararea(int x, int y, int w, int h)
{
  int i;
  for (i = 0; i < h; i++) {
    XDrawLine(display, wmgen.pixmap, BlackGC, x, y+i, x+w, y+i);
  }
}

void openXwindow(int argc, char *argv[], const char * const pixmap_bytes[],
  char *pixmask_bits, int pixmask_width, int pixmask_height)
{
  unsigned int borderwidth = 1;
  XClassHint classHint;
  char *display_name = NULL;
  char *wname = argv[0];
  XTextProperty name;
  XGCValues gcv;
  unsigned long gcm;
  int dummy=0;
  int i;
  for (i=1; argv[i]; i++) {
    if (!strcmp(argv[i], "-display")) 
    display_name = argv[i+1];
  }
  if (!(display = XOpenDisplay(display_name))) {
    fprintf(stderr, "%s: can't open display %s\n", 
      wname, XDisplayName(display_name));
    exit(1);
  }
  screen = DefaultScreen(display);
  Root = RootWindow(display, screen);
  d_depth = DefaultDepth(display, screen);
  x_fd = XConnectionNumber(display);
  GetXPM(&wmgen, pixmap_bytes);
  mysizehints.flags = USSize | USPosition;
  mysizehints.x = 0;
  mysizehints.y = 0;
  back_pix = GetColor("white");
  fore_pix = GetColor("black");
  XWMGeometry(display, screen, Geometry, NULL, borderwidth, &mysizehints,
    &mysizehints.x, &mysizehints.y,&mysizehints.width,&mysizehints.height,
    &dummy);
  mysizehints.width = 64;
  mysizehints.height = 64;
  win = XCreateSimpleWindow(display, Root, mysizehints.x, mysizehints.y,
    mysizehints.width, mysizehints.height, borderwidth, fore_pix, back_pix);
  iconwin = XCreateSimpleWindow(display, win, mysizehints.x, mysizehints.y,
    mysizehints.width, mysizehints.height, borderwidth, fore_pix, back_pix);
  XSetWMNormalHints(display, win, &mysizehints);
  classHint.res_name = wname;
  classHint.res_class = wname;
  XSetClassHint(display, win, &classHint);
  XSelectInput(display, win, ButtonPressMask | ExposureMask |
    ButtonReleaseMask | PointerMotionMask | StructureNotifyMask);
  XSelectInput(display, iconwin, ButtonPressMask | ExposureMask |
    ButtonReleaseMask | PointerMotionMask | StructureNotifyMask);
  if (XStringListToTextProperty(&wname, 1, &name) == 0) {
    fprintf(stderr, "%s: can't allocate window name\n", wname);
    exit(1);
  }
  XSetWMName(display, win, &name);
  gcm = GCForeground | GCBackground | GCGraphicsExposures;
  gcv.graphics_exposures = 0;
  gcv.background = back_pix;
  gcv.foreground = GetColor("#20B2AE");
  NormalGC = XCreateGC(display, Root, gcm, &gcv);
  gcv.foreground = GetColor("#20625E");
  DarkerGC = XCreateGC(display, Root, gcm, &gcv);
  gcv.foreground = GetColor("black");
  BlackGC = XCreateGC(display, Root, gcm, &gcv);
  pixmask = XCreateBitmapFromData(display, win, pixmask_bits,
    pixmask_width, pixmask_height);
  XShapeCombineMask(display, win, ShapeBounding, 0, 0, pixmask, ShapeSet);
  XShapeCombineMask(display, iconwin, ShapeBounding, 0, 0, pixmask, ShapeSet);
  mywmhints.initial_state = WithdrawnState;
  mywmhints.icon_window = iconwin;
  mywmhints.icon_x = mysizehints.x;
  mywmhints.icon_y = mysizehints.y;
  mywmhints.window_group = win;
  mywmhints.flags = StateHint | IconWindowHint | IconPositionHint |
    WindowGroupHint;
  XSetWMHints(display, win, &mywmhints);
  XSetCommand(display, win, argv, argc);
  XMapWindow(display, win);
}
